# -*- coding: utf-8 -*-
from urllib.request import urlopen
from urllib.parse import urlencode
from urllib.error import HTTPError

try:
    import requests
    HAS_REQUEST = True
except ImportError:
    HAS_REQUEST = False

DEFAULT_TIMEOUT = 60

basestring = (str, bytes)

allowed_args = (
    'auth', 'data', 'headers', 'verify',
    'cert', 'config', 'hooks', 'proxies', 'cookies'
)


def _query(url, method, kwargs):
    data = None
    if 'data' in kwargs:
        data = kwargs.pop('data')
    if type(data) in (dict, list, tuple):
        data = urlencode(data)

    if isinstance(method, basestring) and \
       method.lower() == 'get' and data:
        if '?' not in url:
            url += '?'
        elif url[-1] not in ('?', '&'):
            url += '&'
        url += data
        data = None

    if data:
        data = data.encode('utf-8')
    return url, data


def _requests(url, kwargs):

    encoding = kwargs.get('encoding')
    method = kwargs.get('method', 'get').lower()
    session = kwargs.get('session')
    if session:
        meth = getattr(session, str(method))
    else:
        meth = getattr(requests, str(method))
    if method == 'get':
        url, data = _query(url, method, kwargs)
    kw = {}
    for k in allowed_args:
        if k in kwargs:
            kw[k] = kwargs[k]
    resp = meth(url=url, timeout=kwargs.get('timeout', DEFAULT_TIMEOUT), **kw)
    if not (200 <= resp.status_code < 300):
        raise HTTPError(resp.url, resp.status_code,
                        resp.reason, resp.headers, None)
    if encoding:
        resp.encoding = encoding
    html = resp.text
    return html


def _urllib(url, kwargs):
    method = kwargs.get('method')
    url, data = _query(url, method, kwargs)
    return urlopen(url, data, timeout=kwargs.get('timeout', DEFAULT_TIMEOUT))


def url_opener(url, kwargs):
    if HAS_REQUEST:
        return _requests(url, kwargs)
    return _urllib(url, kwargs)



import re


# https://developer.mozilla.org/en-US/docs/Web/HTML/Inline_elements#Elements
INLINE_TAGS = {
    'a', 'abbr', 'acronym', 'b', 'bdo', 'big', 'br', 'button', 'cite',
    'code', 'dfn', 'em', 'i', 'img', 'input', 'kbd', 'label', 'map',
    'object', 'q', 'samp', 'script', 'select', 'small', 'span', 'strong',
    'sub', 'sup', 'textarea', 'time', 'tt', 'var'
}

SEPARATORS = {'br'}


# Definition of whitespace in HTML:
# https://www.w3.org/TR/html4/struct/text.html#h-9.1
WHITESPACE_RE = re.compile(u'[\x20\x09\x0C\u200B\x0A\x0D]+')


def squash_html_whitespace(text):
    # use raw extract_text for preformatted content (like <pre> content or set
    # by CSS rules)
    # apply this function on top of
    return WHITESPACE_RE.sub(' ', text)


def _squash_artifical_nl(parts):
    output, last_nl = [], False
    for x in parts:
        if x is not None:
            output.append(x)
            last_nl = False
        elif not last_nl:
            output.append(None)
            last_nl = True
    return output


def _strip_artifical_nl(parts):
    if not parts:
        return parts
    for start_idx, pt in enumerate(parts):
        if isinstance(pt, str):
            # 0, 1, 2, index of first string [start_idx:...
            break
    iterator = enumerate(parts[:start_idx - 1 if start_idx > 0 else None:-1])
    for end_idx, pt in iterator:
        if isinstance(pt, str):  # 0=None, 1=-1, 2=-2, index of last string
            break
    return parts[start_idx:-end_idx if end_idx > 0 else None]


def _merge_original_parts(parts):
    output, orp_buf = [], []

    def flush():
        if orp_buf:
            item = squash_html_whitespace(''.join(orp_buf)).strip()
            if item:
                output.append(item)
            orp_buf[:] = []

    for x in parts:
        if not isinstance(x, str):
            flush()
            output.append(x)
        else:
            orp_buf.append(x)
    flush()
    return output


def extract_text_array(dom, squash_artifical_nl=True, strip_artifical_nl=True):
    if callable(dom.tag):
        return ''
    r = []
    if dom.tag in SEPARATORS:
        r.append(True)  # equivalent of '\n' used to designate separators
    elif dom.tag not in INLINE_TAGS:
        # equivalent of '\n' used to designate artifically inserted newlines
        r.append(None)
    if dom.text is not None:
        r.append(dom.text)
    for child in dom.getchildren():
        r.extend(extract_text_array(child, squash_artifical_nl=False,
                                    strip_artifical_nl=False))
        if child.tail is not None:
            r.append(child.tail)
    if dom.tag not in INLINE_TAGS and dom.tag not in SEPARATORS:
        # equivalent of '\n' used to designate artifically inserted newlines
        r.append(None)
    if squash_artifical_nl:
        r = _squash_artifical_nl(r)
    if strip_artifical_nl:
        r = _strip_artifical_nl(r)
    return r


def extract_text(dom, block_symbol='\n', sep_symbol='\n', squash_space=True):
    a = extract_text_array(dom, squash_artifical_nl=squash_space)
    if squash_space:
        a = _strip_artifical_nl(_squash_artifical_nl(_merge_original_parts(a)))
    result = ''.join(
        block_symbol if x is None else (
            sep_symbol if x is True else x
        )
        for x in a
    )
    if squash_space:
        result = result.strip()
    return result
